import {
  GraphQLEnumType,
  GraphQLFieldConfig,
  GraphQLFieldConfigArgumentMap,
  GraphQLFieldConfigMap,
  GraphQLFieldResolver,
  GraphQLInputFieldConfig,
  GraphQLInputFieldConfigMap,
  GraphQLInputObjectType,
  GraphQLInputType,
  GraphQLInterfaceType,
  GraphQLNamedType,
  GraphQLObjectType,
  GraphQLOutputType,
  GraphQLScalarType,
  GraphQLSchema,
  GraphQLUnionType,
} from "graphql";
import { NexusArgConfig, NexusArgDef } from "./definitions/args";
import {
  NexusInputFieldDef,
  NexusOutputFieldDef,
} from "./definitions/definitionBlocks";
import { EnumTypeConfig } from "./definitions/enumType";
import {
  NexusExtendTypeConfig,
  NexusExtendTypeDef,
} from "./definitions/extendType";
import { NexusInputObjectTypeConfig } from "./definitions/inputObjectType";
import {
  NexusInterfaceTypeConfig,
  NexusInterfaceTypeDef,
} from "./definitions/interfaceType";
import {
  NexusObjectTypeConfig,
  NexusObjectTypeDef,
} from "./definitions/objectType";
import { NexusScalarTypeConfig } from "./definitions/scalarType";
import { NexusUnionTypeConfig, UnionMembers } from "./definitions/unionType";
import {
  AllNexusInputTypeDefs,
  AllNexusNamedTypeDefs,
} from "./definitions/wrapping";
import {
  GraphQLPossibleInputs,
  GraphQLPossibleOutputs,
  NonNullConfig,
} from "./definitions/_types";
import { TypegenAutoConfigOptions } from "./typegenAutoConfig";
import { TypegenFormatFn } from "./typegenFormatPrettier";
import { GetGen } from "./typegenTypeHelpers";
export declare type Maybe<T> = T | null;
export interface BuilderConfig {
  /**
   * When the schema starts and `process.env.NODE_ENV !== "production"`,
   * artifact files are auto-generated containing the .graphql definitions of
   * the schema
   */
  outputs:
    | {
        /**
         * Absolute path where the GraphQL IDL file should be written
         */
        schema: string | false;
        /**
         * File path where generated types should be saved
         */
        typegen: string | false;
      }
    | false;
  /**
   * Whether the schema & types are generated when the server
   * starts. Default is !process.env.NODE_ENV || process.env.NODE_ENV !== "development"
   */
  shouldGenerateArtifacts?: boolean;
  /**
   * Automatically configure type resolution for the TypeScript
   * representations of the associated types.
   *
   * Alias for typegenConfig: typegenAutoConfig(options)
   */
  typegenAutoConfig?: TypegenAutoConfigOptions;
  /**
   * A configuration function for advanced cases where
   * more control over the `TypegenInfo` is needed.
   */
  typegenConfig?: (
    schema: GraphQLSchema,
    outputPath: string
  ) => TypegenInfo | PromiseLike<TypegenInfo>;
  /**
   * Either an absolute path to a .prettierrc file, or an object
   * with relevant Prettier rules to be used on the generated output
   */
  prettierConfig?: string | object;
  /**
   * Manually apply a formatter to the generated content before saving,
   * see the `prettierConfig` option if you want to use Prettier.
   */
  formatTypegen?: TypegenFormatFn;
  /**
   * Configures the default "nonNullDefaults" for the entire schema the type.
   * Read more about how nexus handles nullability
   */
  nonNullDefaults?: NonNullConfig;
}
export interface TypegenInfo {
  /**
   * Headers attached to the generate type output
   */
  headers: string[];
  /**
   * All imports for the backing types / context
   */
  imports: string[];
  /**
   * A map of all GraphQL types and what TypeScript types they should
   * be represented by.
   */
  backingTypeMap: { [K in GetGen<"objectNames">]?: string };
  /**
   * The type of the context for the resolvers
   */
  contextType?: string;
}
export interface SchemaConfig extends BuilderConfig {
  /**
   * All of the GraphQL types. This is an any for simplicity of developer experience,
   * if it's an object we get the values, if it's an array we flatten out the
   * valid types, ignoring invalid ones.
   */
  types: any;
}
/**
 * Builds all of the types, properly accounts for any using "mix".
 * Since the enum types are resolved synchronously, these need to guard for
 * circular references at this step, while fields will guard for it during lazy evaluation.
 */
export declare class SchemaBuilder {
  protected config: BuilderConfig;
  /**
   * Used to check for circular references.
   */
  protected buildingTypes: Set<any>;
  /**
   * The "final type" map contains all types as they are built.
   */
  protected finalTypeMap: Record<string, GraphQLNamedType>;
  /**
   * The "defined type" map keeps track of all of the types that were
   * defined directly as `GraphQL*Type` objects, so we don't accidentally
   * overwrite any.
   */
  protected definedTypeMap: Record<string, GraphQLNamedType>;
  /**
   * The "pending type" map keeps track of all types that were defined w/
   * Nexus GraphQL and haven't been processed into concrete types yet.
   */
  protected pendingTypeMap: Record<string, AllNexusNamedTypeDefs>;
  /**
   * All "extensions" to types (adding fields on types from many locations)
   */
  protected typeExtensionMap: Record<string, NexusExtendTypeConfig<string>[]>;
  /**
   * Configures the root-level nonNullDefaults defaults
   */
  protected nonNullDefaults: NonNullConfig;
  /**
   * Adds custom dynamic scalar methods to the definition blocks
   * tuple: [FieldName, TypeName]
   */
  protected customScalarMethods: [string, string][];
  constructor(config: BuilderConfig);
  getConfig(): BuilderConfig;
  addType(
    typeDef:
      | AllNexusNamedTypeDefs
      | NexusExtendTypeDef<string>
      | GraphQLNamedType
  ): void;
  getFinalTypeMap(): BuildTypes<any>;
  inputObjectType(
    config: NexusInputObjectTypeConfig<any>
  ): GraphQLInputObjectType;
  objectType(config: NexusObjectTypeConfig<any>): GraphQLObjectType;
  interfaceType(config: NexusInterfaceTypeConfig<any>): GraphQLInterfaceType;
  withScalarMethods<T extends NexusGenCustomScalarMethods<string>>(
    definitionBlock: T
  ): T;
  enumType(config: EnumTypeConfig<any>): GraphQLEnumType;
  unionType(config: NexusUnionTypeConfig<any>): GraphQLUnionType;
  scalarType(config: NexusScalarTypeConfig<string>): GraphQLScalarType;
  protected missingType(typeName: string): GraphQLNamedType;
  protected buildUnionMembers(
    unionName: string,
    members: UnionMembers | undefined
  ): GraphQLObjectType[];
  protected buildObjectFields(
    fields: NexusOutputFieldDef[],
    typeConfig: NexusObjectTypeConfig<any> | NexusInterfaceTypeConfig<any>,
    intoObject: GraphQLFieldConfigMap<any, any>,
    forInterface?: boolean
  ): GraphQLFieldConfigMap<any, any>;
  protected buildInputObjectFields(
    fields: NexusInputFieldDef[],
    typeConfig: NexusInputObjectTypeConfig<string>
  ): GraphQLInputFieldConfigMap;
  protected buildObjectField(
    fieldConfig: NexusOutputFieldDef,
    typeConfig:
      | NexusObjectTypeConfig<string>
      | NexusInterfaceTypeConfig<string>,
    forInterface?: boolean
  ): GraphQLFieldConfig<any, any>;
  protected buildInputObjectField(
    field: NexusInputFieldDef,
    typeConfig: NexusInputObjectTypeConfig<any>
  ): GraphQLInputFieldConfig;
  protected buildArgs(
    args: Record<string, NexusArgDef<string>>,
    typeConfig: NexusObjectTypeConfig<string> | NexusInterfaceTypeConfig<string>
  ): GraphQLFieldConfigArgumentMap;
  protected inputNonNull(
    typeDef:
      | NexusObjectTypeConfig<any>
      | NexusInterfaceTypeConfig<any>
      | NexusInputObjectTypeConfig<any>,
    field: NexusInputFieldDef | NexusArgConfig<any>
  ): boolean;
  protected outputNonNull(
    typeDef: NexusObjectTypeConfig<any> | NexusInterfaceTypeConfig<any>,
    field: NexusOutputFieldDef
  ): boolean;
  protected decorateType<T extends GraphQLNamedType>(
    type: T,
    list: null | undefined | true | boolean[],
    isNonNull: boolean
  ): T;
  protected decorateList<T extends GraphQLOutputType | GraphQLInputType>(
    type: T,
    list: true | boolean[]
  ): T;
  protected getInterface(
    name: string | NexusInterfaceTypeDef<string>
  ): GraphQLInterfaceType;
  protected getEnum(name: string): GraphQLEnumType;
  protected getUnion(name: string): GraphQLUnionType;
  protected getInputObjectType(name: string): GraphQLInputObjectType;
  protected getInputType(
    name: string | AllNexusInputTypeDefs
  ): GraphQLPossibleInputs;
  protected getOutputType(name: string): GraphQLPossibleOutputs;
  protected getObjectType(
    name: string | NexusObjectTypeDef<string>
  ): GraphQLObjectType;
  protected getOrBuildType(
    name: string | AllNexusNamedTypeDefs
  ): GraphQLNamedType;
  protected getResolver(
    fieldOptions: NexusOutputFieldDef,
    typeConfig: NexusObjectTypeConfig<any> | NexusInterfaceTypeConfig<any>,
    forInterface?: boolean
  ):
    | GraphQLFieldResolver<
        any,
        any,
        {
          [argName: string]: any;
        }
      >
    | undefined;
  missingResolveType(name: string, location: "union" | "interface"): () => null;
}
export interface BuildTypes<
  TypeMapDefs extends Record<string, GraphQLNamedType>
> {
  typeMap: TypeMapDefs;
}
/**
 * Builds the types, normalizing the "types" passed into the schema for a
 * better developer experience. This is primarily useful for testing
 * type generation
 */
export declare function buildTypes<
  TypeMapDefs extends Record<string, GraphQLNamedType> = any
>(
  types: any,
  config?: BuilderConfig,
  schemaBuilder?: SchemaBuilder
): BuildTypes<TypeMapDefs>;
/**
 * Builds the schema, we may return more than just the schema
 * from this one day.
 */
export declare function makeSchemaInternal(
  options: SchemaConfig,
  schemaBuilder?: SchemaBuilder
): {
  schema: GraphQLSchema;
};
/**
 * Defines the GraphQL schema, by combining the GraphQL types defined
 * by the Nexus GraphQL layer or any manually defined GraphQLType objects.
 *
 * Requires at least one type be named "Query", which will be used as the
 * root query type.
 */
export declare function makeSchema(options: SchemaConfig): GraphQLSchema;
