"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var graphql_1 = require("graphql");
var utils_1 = require("./utils");
function convertSDL(sdl, commonjs, json) {
    if (commonjs === void 0) { commonjs = false; }
    if (json === void 0) { json = JSON; }
    try {
        return new SDLConverter(sdl, commonjs, json).print();
    }
    catch (e) {
        return "Error Parsing SDL into Schema: " + e.stack;
    }
}
exports.convertSDL = convertSDL;
/**
 * Convert an existing SDL schema into a Nexus GraphQL format
 */
var SDLConverter = /** @class */ (function () {
    function SDLConverter(sdl, commonjs, json) {
        if (commonjs === void 0) { commonjs = false; }
        if (json === void 0) { json = JSON; }
        this.json = json;
        this.export =
            commonjs === null ? "const " : commonjs ? "exports." : "export const ";
        this.schema = graphql_1.buildSchema(sdl);
        this.groupedTypes = utils_1.groupTypes(this.schema);
    }
    SDLConverter.prototype.print = function () {
        return [
            this.printObjectTypes(),
            this.printInterfaceTypes(),
            this.printInputObjectTypes(),
            this.printUnionTypes(),
            this.printEnumTypes(),
            this.printScalarTypes(),
        ]
            .filter(function (f) { return f; })
            .join("\n\n");
    };
    SDLConverter.prototype.printObjectTypes = function () {
        var _this = this;
        if (this.groupedTypes.object.length > 0) {
            return this.groupedTypes.object
                .map(function (t) { return _this.printObjectType(t); })
                .join("\n");
        }
        return "";
    };
    SDLConverter.prototype.printObjectType = function (type) {
        var _this = this;
        var implementing = type.getInterfaces().map(function (i) { return i.name; });
        var implementsInterfaces = implementing.length > 0
            ? "    t.implements(" + implementing
                .map(function (i) { return _this.json.stringify(i); })
                .join(", ") + ")"
            : "";
        return this.printBlock([
            "" + this.export + type.name + " = objectType({",
            "  name: \"" + type.name + "\",",
            "  definition(t) {",
            implementsInterfaces,
            this.printObjectFields(type),
            "  }",
            "})",
        ]);
    };
    SDLConverter.prototype.printObjectFields = function (type) {
        var _this = this;
        return utils_1.objValues(type.getFields())
            .map(function (field) {
            if (graphql_1.isObjectType(type) && utils_1.isInterfaceField(type, field.name)) {
                return;
            }
            return _this.printField("output", field);
        })
            .filter(function (f) { return f; })
            .join("\n");
    };
    SDLConverter.prototype.printField = function (source, field) {
        var fieldType = field.type;
        var isNonNull = false;
        var list = [];
        while (graphql_1.isWrappingType(fieldType)) {
            while (graphql_1.isListType(fieldType)) {
                fieldType = fieldType.ofType;
                if (graphql_1.isNonNullType(fieldType)) {
                    fieldType = fieldType.ofType;
                    list.unshift(true);
                }
                else {
                    list.unshift(false);
                }
            }
            if (graphql_1.isNonNullType(fieldType)) {
                isNonNull = true;
                fieldType = fieldType.ofType;
            }
        }
        var prefix = list.length === 1 ? "t.list." : "t.";
        return "    " + prefix + this.printFieldMethod(source, field, fieldType, list, isNonNull);
    };
    SDLConverter.prototype.printFieldMethod = function (source, field, type, list, isNonNull) {
        var objectMeta = {};
        if (field.description) {
            objectMeta.description = field.description;
        }
        if (field.deprecationReason) {
            objectMeta.deprecation = field.deprecationReason;
        }
        if (list.length > 1) {
            objectMeta.list = list;
        }
        if (!isNonNull && source === "output") {
            objectMeta.nullable = true;
        }
        else if (isNonNull && source === "input") {
            objectMeta.required = true;
        }
        var str = "";
        if (isCommonScalar(type)) {
            str += type.name.toLowerCase() + "(\"" + field.name + "\"";
        }
        else {
            objectMeta.type = type;
            str += "field(\"" + field.name + "\"";
        }
        if (Object.keys(objectMeta).length > 0) {
            str += ", " + this.json.stringify(objectMeta);
        }
        return str + ")";
    };
    SDLConverter.prototype.printInterfaceTypes = function () {
        var _this = this;
        if (this.groupedTypes.interface.length) {
            return this.groupedTypes.interface
                .map(function (t) { return _this.printInterfaceType(t); })
                .join("\n");
        }
        return "";
    };
    SDLConverter.prototype.printInterfaceType = function (type) {
        return this.printBlock([
            "" + this.export + type.name + " = interfaceType({",
            "  name: \"" + type.name + "\",",
            this.maybeDescription(type),
            "  definition(t) {",
            this.printObjectFields(type),
            "  }",
            "});",
        ]);
    };
    SDLConverter.prototype.printEnumTypes = function () {
        var _this = this;
        if (this.groupedTypes.enum.length) {
            return this.groupedTypes.enum
                .map(function (t) { return _this.printEnumType(t); })
                .join("\n");
        }
        return "";
    };
    SDLConverter.prototype.printEnumType = function (type) {
        var members = type.getValues().map(function (val) {
            var description = val.description, name = val.name, deprecationReason = val.deprecationReason, value = val.value;
            if (!description && !deprecationReason && name === value) {
                return val.name;
            }
            return { description: description, name: name, deprecated: deprecationReason, value: value };
        });
        return this.printBlock([
            "" + this.export + type.name + " = enumType({",
            "  name: \"" + type.name + "\",",
            this.maybeDescription(type),
            "  members: " + this.json.stringify(members) + ",",
            "});",
        ]);
    };
    SDLConverter.prototype.printInputObjectTypes = function () {
        var _this = this;
        if (this.groupedTypes.input.length) {
            return this.groupedTypes.input
                .map(function (t) { return _this.printInputObjectType(t); })
                .join("\n");
        }
        return "";
    };
    SDLConverter.prototype.printInputObjectType = function (type) {
        return this.printBlock([
            "" + this.export + type.name + " = inputObjectType({",
            "  name: \"" + type.name + "\",",
            this.maybeDescription(type),
            "  definition(t) {",
            "  }",
            "});",
        ]);
    };
    SDLConverter.prototype.printUnionTypes = function () {
        var _this = this;
        if (this.groupedTypes.union.length) {
            return this.groupedTypes.union
                .map(function (t) { return _this.printUnionType(t); })
                .join("\n");
        }
        return "";
    };
    SDLConverter.prototype.printUnionType = function (type) {
        return this.printBlock([
            "" + this.export + type.name + " = unionType({",
            "  name: \"" + type.name + "\",",
            this.maybeDescription(type),
            "  definition(t) {",
            "  }",
            "});",
        ]);
    };
    SDLConverter.prototype.printScalarTypes = function () {
        var _this = this;
        if (this.groupedTypes.scalar.length) {
            return this.groupedTypes.scalar
                .filter(function (s) { return !graphql_1.isSpecifiedScalarType(s); })
                .map(function (t) { return _this.printScalarType(t); })
                .join("\n");
        }
        return "";
    };
    SDLConverter.prototype.printScalarType = function (type) {
        return this.printBlock([
            "" + this.export + type.name + " = scalarType({",
            "  name: \"" + type.name + "\",",
            this.maybeDescription(type),
            this.maybeAsNexusType(type),
            "  serialize() { /* Todo */ },",
            "  parseValue() { /* Todo */ },",
            "  parseLiteral() { /* Todo */ }",
            "});",
        ]);
    };
    SDLConverter.prototype.maybeAsNexusType = function (type) {
        if (isCommonScalar(type)) {
            return "  asNexusMethod: \"" + type.name.toLowerCase() + "\",";
        }
        return null;
    };
    SDLConverter.prototype.maybeDescription = function (type) {
        if (type.description) {
            return "  description: " + JSON.stringify(type.description) + ",";
        }
        return null;
    };
    SDLConverter.prototype.printBlock = function (block) {
        return block.filter(function (t) { return t !== null && t !== ""; }).join("\n");
    };
    return SDLConverter;
}());
exports.SDLConverter = SDLConverter;
function isCommonScalar(field) {
    if (graphql_1.isScalarType(field)) {
        return (graphql_1.isSpecifiedScalarType(field) ||
            field.name === "UUID" ||
            field.name === "Date");
    }
    return false;
}
//# sourceMappingURL=sdlConverter.js.map